;DCF77 simulator with keypad, LCD and RS232 port
;ATTiny2313 @ 4.194304 MHz (external XTAL), Fuses: low=0xDD, high=0x9B

;PORTS:
;  PB0 = Keypad column 1,4,7,*
;  PB1 = Keypad comumn 2,5,8,0
;  PB2 = Keypad column 3,6,9,#
;  PB3 = [unused]
;  PB4 = LCD D4, Keypad row 1,2,3
;  PB5 = LCD D5, Keypad row 4,5,6
;  PB6 = LCD D6, Keypad row 7,8,9
;  PB7 = LCD D7, Keypad row *,0,#
;  PD0 = RS232 Rx
;  PD1 = RS232 Tx
;  PD2 = [unused]
;  PD3 = [unused]
;  PD4 = LCD RS
;  PD5 = LCD E
;  PD6 = DCF77 signal output

;VERSIONS
;  2011-03-12  Arne Rossius
;    * first version

;registers for DCF77 frame:
;         MSB                                       LSB
;       +-----+-----+-----+-----+-----+-----+-----+-----+   CA = Call bit
;dcf1 = | m2  | m1  |  S  | LS  | WIN | SUM | S/W | CA  |   S/W = Switch DST
;       +-----+-----+-----+-----+-----+-----+-----+-----+   SUM = summer time
;dcf2 = | h2  | h1  | mP  | m40 | m20 | m10 | m8  | m4  |   WIN = winter time
;       +-----+-----+-----+-----+-----+-----+-----+-----+   LS = leap second
;dcf3 = | D4  | D2  | D1  | hP  | h20 | h10 | h8  | h4  |   S = start bit (1)
;       +-----+-----+-----+-----+-----+-----+-----+-----+   m = minute
;dcf4 = | M2  | M1  | WD4 | WD2 | WD1 | D20 | D10 | D8  |   h = hour
;       +-----+-----+-----+-----+-----+-----+-----+-----+   D = day of month
;dcf5 = | Y10 | Y8  | Y4  | Y2  | Y1  | M10 | M8  | M4  |   W = day of week
;       +-----+-----+-----+-----+-----+-----+-----+-----+   M = month
;dcf6 = |  0  |  0  |  0  |  0  | DaP | Y80 | Y40 | Y20 |   Y = year
;       +-----+-----+-----+-----+-----+-----+-----+-----+   *P = parity bit

;TODO: Send "empty" bit after setting new date/time
;TODO: Set time/date and weekday in same screen?
;TODO: remove extra "enter" after minute ones
;TODO: jump to minute ones when pressing backspace for weekday

.include "../include/tn2313def.inc"

.def	dcf1 = R0
.def	dcf2 = R1
.def	dcf3 = R2
.def	dcf4 = R3
.def	dcf5 = R4
.def	dcf6 = R5
.def	zero = R6
.def	oldkey = R7
.def	sreg_backup = R8

.def	temp = R16
.def	temp2 = R17
.def	second = R18
.def	subsec = R19
.def	key = R20
.def	buflen = R21
.def	w1 = R22
.def	w2 = R23

;R28 = Y pointer (for RAM access, only low byte needed)
.def	flags = R29
	.equ	fUpdateTime = 0
	.equ	fCall = 1
	.equ	fLeapSecond = 2
	.equ	fLeapSecondNow = 3
	.equ	fDCFBit = 4
	.equ	fLowerNibble = 5
;R31:R30 = Z pointer (for Flash access)

;keypad key definitions
;keys '0' to '9' = 0 ~ 9
.equ	kASTERISK = 0x10
.equ	kHASH = 0x11
.equ	kNONE = 0xFF

;weekday definitions
.equ	MONDAY = 1
.equ	TUESDAY = 2
.equ	WEDNESDAY = 3
.equ	THURSDAY = 4
.equ	FRIDAY = 5
.equ	SATURDAY = 6
.equ	SUNDAY = 7

;===============================================================================

.macro _brne
	breq	PC+2
	rjmp	@0
.endmacro

.macro _breq
	brne	PC+2
	rjmp	@0
.endmacro

.macro _brts
	brtc	PC+2
	rjmp	@0
.endmacro

;===============================================================================

.dseg
	;7 bytes = (minutes, hours, weekday, day of month, month, year, flags)
	RAM_Times_Start:
	RAM_DCFTime: .byte 7 ;time for DCF77 frame (real time + 1 minute)
	RAM_Time: .byte 7 ;real time
	RAM_NewTime: .byte 7 ;new time (from user entry)
	RAM_Times_End:
	
	RAM_Buffer: .byte 32 ;RS232 input buffer

;===============================================================================

.cseg
.org 0x000
	;init stackpointer
	ldi	temp, RAMEND
	out	SPL, temp
	
	;init ports
	ldi	temp, 0xF0
	out	DDRB, temp
	;(continued below)
	rjmp	reset

.org OVF1addr
	rjmp	ovf1

.org OVF0addr
	rjmp	ovf0
	
.include "lcd.asm"
.include "clock.asm"
.include "uart.asm"
.include "eeprom.asm"

;===============================================================================

.eseg

keypad_table:
	.db 0xEF, 7, 8, 9, kNONE, 0xDF, 4, 5, 6, kNONE
	.db 0xBF, 1, 2, 3, kNONE, 0x7F, kASTERISK, 0, kHASH, kNONE
	
weekdays:
	.db "Mo", "Tu", "We", "Th", "Fr", "Sa", "Su"
	
str_new_date_time:
	.db "New Date/Time:  ",0
str_invalid_setting:
	.db "Invalid setting!",0
str_new_weekday_1:
	.db "New Weekday: ",0
str_new_weekday_2:
	.db "1=Mo ... 7=Su",0
str_summertime:
	.db "Summer Time?",0
str_yes_no:
	.db "1=YES, 2=NO",0
time_setting:
	;start-up time/date setting
	.org time_setting+tYear
		.db 0x42
	.org time_setting+tMonth
		.db 0x05
	.org time_setting+tDay
		.db 0x23
	.org time_setting+tHour
		.db 0x12
	.org time_setting+tMinute
		.db 0x34
	.org time_setting+tWeekday
		.db FRIDAY
	.org time_setting+tFlags
		.db 1<<timeDST ;(winter time: 0, summer time: 1<<timeDST)

.cseg

;===============================================================================

ovf1:
	;Timer 1 overflow: DCF bit length counter reached "zero"
	cbi	PORTD, 5
	sbi	PORTD, 6
	out	TCCR1B, zero ;timer stopped
	reti

;===============================================================================

ovf0:
	;Timer 0 overflow interrupt (64 Hz periodic interrupt)
	in	sreg_backup, SREG
	push	temp
	push	temp2
	push	YL
	push	ZL
	
	;poll keypad
	ldi	temp, kNONE
	ldi	ZL, keypad_table
	ldi	YL, 0
ovf0_keypad_loop:
	push	temp
	rcall	ee_read
	out	PORTB, temp
	pop	temp
	lpm	temp2, Z ;3 cycles delay
	in	temp2, PINB
ovf0_keypad_bitloop:
	sbrs	temp2, 0
	rcall	ee_read
	sbrc	temp2, 0
	inc	ZL
	lsr	temp2
	subi	YL, -0x04 ;halfcarry cleared after 4x "sub -0x04"
	brhs	ovf0_keypad_bitloop
	sbrs	YL, 6 ;bit 6 set after 4 inner loops
	rjmp	ovf0_keypad_loop
	;check if key was pressed already
	cp	temp, oldkey
	breq	ovf0_keypad_end
	mov	oldkey, temp
	cpi	temp, kNONE
	breq	ovf0_keypad_end
	mov	key, temp
ovf0_keypad_end:
	
	;divide to 1 Hz
	inc	subsec
	cpi	subsec, 64
	brlo	ovf0_end
	
	;increment seconds
	clr	subsec
	mov	temp, second
	rcall	bcdinc
	mov	second, temp
	cpi	temp, 0x60
	sbrc	flags, fLeapSecondNow
	cpi	temp, 0x61
	brlo	ovf0_inctime_end
	
	;increment time by 1 minute
	cpi	temp, 0x61
	brlo	PC+2
	andi	flags, ~(1<<fLeapSecond | 1<<fLeapSecondNow)
	clr	second
	ldi	YL, RAM_Time
	rcall	inctime
	rcall	update_dcftime
	
ovf0_inctime_end:
	ori	flags, 1<<fUpdateTime

	;set DCF77 output
	rcall	dcf_output
	
ovf0_end:
	pop	ZL
	pop	YL
	pop	temp2
	pop	temp
	out	SREG, sreg_backup
	reti

;===============================================================================

reset:
	;init ports (continued from above)
	ldi	temp, 0x0F
	out	PORTB, temp
	ldi	temp, 0x7E
	out	DDRD, temp
	ldi	temp, 0x43
	out	PORTD, temp
	
	;init UART
	ldi	temp, 1<<TXEN | 1<<RXEN
	out	UCSRB, temp
	ldi	temp, 26
	out	UBRRL, temp
	
	;init registers
	clr	zero
	clr	second
	clr	subsec
	clr	buflen
	ldi	flags, 1<<fUpdateTime
	ldi	key, kNONE
	mov	oldkey, key
	
	;clear time/date RAM
	ldi	YL, RAM_Times_Start
init_ram:
	st	Y+, zero
	cpi	YL, RAM_Times_End
	brne	init_ram
	
	;init timer for clock
	ldi	temp, 0x04 ;Clk/256
	out	TCCR0B, temp
	
	;enable timer overflow interrupts
	ldi	temp, 1<<TOIE0 | 1<<TOIE1
	out	TIMSK, temp
	
	;set initial time/date
	ldi	YL, RAM_Time
	rcall	load_time
	ldi	YL, RAM_Time
	rcall	check_dst ;TODO necessary? (information in EEPROM should be correct)
	rcall	update_dcftime
	rcall	dcf_output
	
	;enable interrupts
	sei
	
	;init LCD
	rcall	lcd_init
	
;-------------------------------------------------------------------------------

loop:
	;poll UART
	sbic	UCSRA, RXC
	rcall	uart_rx
	
	;update LCD when requested
	sbrc	flags, fUpdateTime
	rcall	display_main_screen
	
	;check for keypress
	cpi	key, 1
	breq	set_time
	cpi	key, 2
	breq	set_leapsec
	cpi	key, 3
	breq	set_call
	
	rjmp	loop
	
;--------------------

set_leapsec:
	;toggle leap-second at end of hour on/off
	ldi	temp, 1<<fLeapSecond
	eor	flags, temp
	ldi	key, kNONE
	ori	flags, 1<<fUpdateTime
	rjmp	loop

;--------------------

set_call:
	;toggle Call Bit on/off
	ldi	temp, 1<<fCall
	eor	flags, temp
	ldi	key, kNONE
	ori	flags, 1<<fUpdateTime
	rjmp	loop

;--------------------

set_time:
	;set new time

	;use last setting as default (read from EEPROM)
	ldi	YL, RAM_NewTime
	rcall	load_time
	
set_time_restart:
	;display message and current setting
	rcall	lcd_clear
	_lcd_eestr str_new_date_time
	ldi	YL, RAM_NewTime
	rcall	display_time
	ldi	temp, 0x0D ;blinking cursor on
	rcall	lcd_cmd
	andi	flags, ~(1<<fLowerNibble)
	
set_year:
	;year
	ldi	YL, RAM_NewTime+tYear
	_lcd_goto 2, 1
	rcall	set_bcd
	cpi	key, kASTERISK
	_breq	set_time_end
	
set_month:
	;month
	ldi	YL, RAM_NewTime+tMonth
	_lcd_goto 5, 1
	rcall	set_bcd
	cpi	key, kASTERISK
	breq	set_year
	
set_day:
	;day of month
	ldi	YL, RAM_NewTime+tDay
	_lcd_goto 8, 1
	rcall	set_bcd
	cpi	key, kASTERISK
	breq	set_month
	
set_hour:
	;hour
	ldi	YL, RAM_NewTime+tHour
	_lcd_goto 11, 1
	rcall	set_bcd
	cpi	key, kASTERISK
	breq	set_day
	
set_minute:
	;minute
	ldi	YL, RAM_NewTime+tMinute
	_lcd_goto 14, 1
	rcall	set_bcd
	cpi	key, kASTERISK
	breq	set_hour
	
	;wait for backspace or enter
	ori	flags, 1<<fLowerNibble
	_lcd_goto 16, 1
set_done:
	rcall	waitkey
	cpi	key, kASTERISK
	breq	set_minute
	cpi	key, kHASH
	brne	set_done
	
	;date/time confirmed, check for validity
	clt
	rcall	check_newtime
	brtc	set_weekday
	;date/time invalid
	_lcd_goto 0, 0
	_lcd_eestr str_invalid_setting
set_invalid_wait:
	rcall	waitkey
	cpi	key, 10
	brlo	set_invalid_wait
	_lcd_goto 0, 0
	_lcd_eestr str_new_date_time
	cpi	key, kASTERISK
	breq	set_minute
	rjmp	set_done
	
set_weekday:
	;ask for weekday
	rcall	lcd_clear
	ldi	temp, 0x0C ;cursor off
	rcall	lcd_cmd
	_lcd_eestr str_new_weekday_1
	ldi	YL, RAM_NewTime
	rcall	display_weekday
	_lcd_goto 0, 1
	_lcd_eestr str_new_weekday_2
set_weekday_loop:
	rcall	waitkey
	cpi	key, kHASH
	breq	set_weekday_end
	cpi	key, kASTERISK
	_breq	set_time_restart ;go back to date/time setting
	cpi	key, 0
	breq	set_weekday_loop
	cpi	key, 8
	brsh	set_weekday_loop
	std	Y+tWeekday, key
	_lcd_goto 13, 0
	rcall	display_weekday
	rjmp	set_weekday_loop
set_weekday_end:
	
set_dst:
	;check summertime, ask if ambiguous
	ldi	YL, RAM_NewTime
	rcall	check_dst
	cpi	temp, DST_STARTS
	breq	set_dst_invalid
	cpi	temp, DST_YES
	breq	set_dst_yes
	cpi	temp, DST_UNKNOWN
	breq	set_dst_ask
set_dst_no:
	;winter time (DST_NO or DST_ENDS)
	ldd	temp, Y+tFlags
	andi	temp, ~(1<<timeDST)
	std	Y+tFlags, temp
	rjmp	set_dst_end
set_dst_invalid:
	;02:xx on last sunday in march: time doesn't exist, switch to summer time
	ldi	temp, 0x03
	std	Y+tHour, temp
	;(fall through to set_dst_yes)
set_dst_yes:
	;summer time (DST_YES)
	ldd	temp, Y+tFlags
	ori	temp, 1<<timeDST
	std	Y+tFlags, temp
	rjmp	set_dst_end
set_dst_ask:
	;ask if current time is summer or winter time (DST_UNKNOWN)
	rcall	lcd_clear
	_lcd_eestr str_summertime
	_lcd_goto 0, 1
	_lcd_eestr str_yes_no
set_dst_key:
	rcall	waitkey
	cpi	key, 1
	breq	set_dst_yes
	cpi	key, 2
	breq	set_dst_no
	rjmp	set_dst_key
set_dst_end:
	rcall	check_dst
	
	;set new time/date
	rcall	set_newtime
	
set_time_end:
	rcall	lcd_clear
	ldi	key, kNONE
	ori	flags, 1<<fUpdateTime
	rjmp	loop
	
;===============================================================================

set_bcd:
	;set packed BCD value
	sbrs	flags, fLowerNibble
	rjmp	set_bcd_1
set_bcd_1to2:
	ldi	temp, 0x14 ;move cursor to the right
	rcall	lcd_cmd
	rjmp	set_bcd_2
set_bcd_2to1:
	;move cursor from low to high nibble position
	ldi	temp, 0x10 ;move cursor to the left
	rcall	lcd_cmd
set_bcd_1:
	;enter high nibble
	rcall	waitkey
	cpi	key, kASTERISK
	breq	set_bcd_cancel
	cpi	key, kHASH
	breq	set_bcd_1to2
	ld	temp, Y
	andi	temp, 0x0F
	mov	temp2, key
	swap	temp2
	or	temp, temp2
	st	Y, temp
	rcall	set_bcd_disp
set_bcd_2:
	;enter low nibble
	ori	flags, 1<<fLowerNibble
	rcall	waitkey
	cpi	key, kASTERISK
	breq	set_bcd_2to1
	cpi	key, kHASH
	breq	set_bcd_end
	ld	temp, Y
	andi	temp, 0xF0
	or	temp, key
	st	Y+, temp
	;(fall through to set_bcd_disp)
set_bcd_disp:
	mov	temp, key
	ori	temp, '0'
	rcall	lcd_chr
set_bcd_end:
	andi	flags, ~(1<<fLowerNibble)
	ret
set_bcd_cancel:
	ori	flags, 1<<fLowerNibble
	ret

;--------------------

waitkey:
	;wait for keypress
	ldi	key, kNONE
waitkey_loop:
	cpi	key, kNONE
	breq	waitkey_loop
	ret

;===============================================================================

.macro _compare
	ldi	temp2, @0
	rcall	uart_compare
.endmacro

uart_rx:
	;receive a byte from UART
	in	temp, UDR
	cpi	temp, 0x20 ;printable char
	brsh	uart_buffer
	cpi	temp, 0x0a ;LF
	breq	uart_eval
	cpi	temp, 0x0d ;CR
	breq	uart_eval
	ret
	
	
	
uart_buffer:
	;store char to buffer
	cpi	buflen, 32
	brsh	uart_buffer_end ;buffer full
	cpi	temp, 'A'
	brlo	uart_buffer_st
	cpi	temp, 'Z'+1
	brsh	uart_buffer_st
	ori	temp, 0x20 ;convert to lowercase (set bit 5)
uart_buffer_st:
	ldi	YL, RAM_Buffer
	add	YL, buflen
	st	Y, temp
	inc	buflen
uart_buffer_end:
	ret
	
	
	
uart_eval:
	;evaluate buffer contents
	cpi	buflen, 0
	breq	uart_eval_end
	;check buffer contents
	ldi	YL, RAM_Buffer
	clt
	cpi	buflen, 1
	_breq	uart_dump
	cpi	buflen, 4
	breq	uart_cmd
	cpi	buflen, 23
	breq	uart_time
	rjmp	uart_error

uart_eval_end:
	ldi	buflen, 0
	ret
	


uart_cmd:
	;command ("LS=x" or "CA=x")
	ld	temp, Y+
	cpi	temp, 'l'
	breq	uart_ls
	cpi	temp, 'c'
	breq	uart_ca
	rjmp	uart_error
	
;--------------------
uart_set:
	or	flags, temp
	rjmp	uart_ok
	
uart_clear:
	com	temp
	and	flags, temp
	rjmp	uart_ok
;--------------------

uart_ls:
	;enable/disable leap second at end of hour
	_compare 's'
	_compare '='
	_brts	uart_error
	ldi	temp, 1<<fLeapSecond
	rjmp	uart_set_flag
	
uart_ca:
	;set/reset Call Bit
	_compare 'a'
	_compare '='
	brts	uart_error
	ldi	temp, 1<<fCall
	;(fall through to uart_set_flag)
	
uart_set_flag:
	;set flag when byte is '1', clear it when byte is '0'
	ld	temp2, Y+
	cpi	temp2, '1'
	breq	uart_set
	cpi	temp2, '0'
	breq	uart_clear
	rjmp	uart_error
	
	
	
uart_time:
	;set time/date, format: T=20YY-MM-DD hh:mm+Z WD
	ldi	ZL, RAM_NewTime
	_compare 't'
	_compare '='
	;century (must be 20)
	rcall	uart_bcd
	cpi	temp, 0x20
	brne	uart_error
	;year
	rcall	uart_bcd
	std	Z+tYear, temp
	_compare '-'
	;month
	rcall	uart_bcd
	std	Z+tMonth, temp
	_compare '-'
	;day
	rcall	uart_bcd
	std	Z+tDay, temp
	_compare ' '
	;hour
	rcall	uart_bcd
	std	Z+tHour, temp
	_compare ':'
	;minute
	rcall	uart_bcd
	std	Z+tMinute, temp
	_compare '+'
	;timezone (summer or winter time)
	ldi	temp2, 0
	ld	temp, Y+
	cpi	temp, '1'
	breq	PC+4
	cpi	temp, '2'
	brne	uart_error
	ldi	temp2, 1<<timeDST
	std	Z+tFlags, temp2
	_compare ' '
	;weekday
	ldi	ZL, weekdays
	ld	w1, Y+
	ld	w2, Y+
uart_weekday_loop:
	rcall	ee_read
	ori	temp, 0x20 ;convert to lowercase (set bit 5)
	mov	temp2, temp
	rcall	ee_read
	cp	temp2, w1
	cpc	temp, w2
	breq	uart_weekday_found
	cpi	ZL, weekdays+14
	brlo	uart_weekday_loop
	;(fall through to uart_error)

uart_error:
uart_error_dst:
	;error (syntax, range or DST), send "ERROR"
	rcall	putstr
		.db "ERROR",0x0d,0x0a,0
	rjmp	uart_eval_end
	
uart_weekday_found:
	subi	ZL, weekdays
	lsr	ZL
	ldi	YL, RAM_NewTime
	std	Y+tWeekday, ZL
	;check time/date
	rcall	check_newtime
	brts	uart_error
	;check DST, use given value on last sunday in october 02:xx
	rcall	check_dst
	ldd	temp2, Y+tFlags
	cpi	temp, DST_STARTS
	breq	uart_error_dst
	cpi	temp, DST_YES
	breq	uart_dst_yes
	cpi	temp, DST_UNKNOWN
	breq	uart_dst_end ;use value given in received string
	;winter time
	sbrc	temp2, timeDST
	rjmp	uart_error_dst
	rjmp	uart_dst_end
uart_dst_yes:
	;summer time
	sbrs	temp2, timeDST
	rjmp	uart_error_dst
	rjmp	uart_dst_end
uart_dst_end:
	;set time/date to new values
	rcall	set_newtime
	;(fall through to uart_ok)
	
	
	
uart_ok:
	ori	flags, 1<<fUpdateTime
	rcall	putstr
		.db "OK",0x0d,0x0a,0
	rjmp	uart_eval_end
	
	
	
uart_dump:
	;dump current time/date
	ld	temp, Y
	cpi	temp, '?'
	brne	uart_error
	_putstr "T=20"
	;year
	ldi	YL, RAM_Time
	ldd	temp, Y+tYear
	rcall	putbcd
	_putc '-'
	;month
	ldi	YL, RAM_Time
	ldd	temp, Y+tMonth
	rcall	putbcd
	_putc '-'
	;day
	ldi	YL, RAM_Time
	ldd	temp, Y+tDay
	rcall	putbcd
	_putc ' '
	;hour
	ldi	YL, RAM_Time
	ldd	temp, Y+tHour
	rcall	putbcd
	_putc ':'
	;minute
	ldi	YL, RAM_Time
	ldd	temp, Y+tMinute
	rcall	putbcd
	_putc ':'
	;seconds
	mov	temp, second
	rcall	putbcd
	_putc '+'
	;timezone
	ldi	temp, '1'
	ldd	temp2, Y+tFlags
	sbrc	temp2, timeDST
	ldi	temp, '2'
	rcall	putc
	_putc ' '
	;weekday
	ldd	ZL, Y+tWeekday
	lsl	ZL
	subi	ZL, 2-weekdays
	rcall	ee_read
	rcall	putc
	rcall	ee_read
	rcall	putc
	rcall	putstr
		.db 0x0d, 0x0a, 0
	rjmp	uart_eval_end

;--------------------

uart_compare:
	;compare char from RAM with char in temp2, set T flag on mismatch
	ld	temp, Y+
	cp	temp, temp2
	breq	PC+2
	set
	ret

;--------------------

uart_bcd:
	;get packed BCD number from RAM (2 digits), set T flag on error
	ld	temp, Y+
	rcall	check_digit
	swap	temp
	mov	temp2, temp
	ld	temp, Y+
	rcall	check_digit
	or	temp, temp2
	ret
	
;--------------------
	
check_digit:
	;check ASCII digit, return binary digit if valid, set T flag on error
	cpi	temp, '0'
	brlo	check_failed
	cpi	temp, '9'+1
	brsh	check_failed
	andi	temp, 0x0F
	ret
	
;===============================================================================

check_newtime:
	;check new time/date for validity, set T flag when failed
	ldi	YL, RAM_NewTime
	ldd	temp, Y+tMinute
	cpi	temp, 0x60
	brsh	check_failed
	ldd	temp, Y+tHour
	cpi	temp, 0x24
	brsh	check_failed
	ldd	temp, Y+tMonth
	cpi	temp, 0x00
	breq	check_failed
	cpi	temp, 0x13
	brsh	check_failed
	rcall	calculate_month_length ;result goes into temp2
	ldd	temp, Y+tDay
	cpi	temp, 0x00
	breq	check_failed
	cp	temp2, temp
	brlo	check_failed
	ret
	
;--------------------
	
check_failed:
	;check failed, set T flag to indicate error
	set
	ret
	
;--------------------

set_newtime:
	;set current time/date to new time/date setting
	
	;copy new time/date to current time/date and to EEPROM
	clr	subsec
	clr	second
	ldi	YL, RAM_Time
	ldi	ZL, time_setting
set_newtime_copy:
	ldd	temp, Y+(RAM_NewTime-RAM_Time)
	st	Y+, temp
	rcall	ee_write
	cpi	YL, RAM_Time+7
	brne	set_newtime_copy
	
	;update DCF77 data
	rcall	update_dcftime
	rjmp	dcf_output ;combined "rcall dcf_output" and "ret"
	
;===============================================================================

load_time:
	;copy time from EEPROM to RAM[Y]
	ldi	temp2, 7
	ldi	ZL, time_setting
load_time_loop:
	rcall	ee_read
	st	Y+, temp
	dec	temp2
	brne	load_time_loop
	ret

;--------------------

display_main_screen:
	;display weekday, settings, second, date and time
	andi	flags, ~(1<<fUpdateTime)
	ldi	YL, RAM_Time
	;weekday
	_lcd_goto 0, 0
	rcall	display_weekday
	_lcd_chr ' '
	;summer/winter time
	ldd	temp2, Y+tFlags
	ldi	temp, 'W'
	sbrc	temp2, timeDST
	ldi	temp, 'S'
	rcall	lcd_chr
	ldd	temp2, Y+tFlags
	ldi	temp, ' '
	sbrc	temp2, timeAnnounceDST
	ldi	temp, '*'
	rcall	lcd_chr
	_lcd_chr ' '
	;leap second
	ldi	temp, ' '
	sbrc	flags, fLeapSecond
	ldi	temp, 'L'
	rcall	lcd_chr
	ldi	temp, ' '
	sbrc	flags, fLeapSecond
	ldi	temp, 'S'
	rcall	lcd_chr
	_lcd_chr ' '
	;call bit
	ldi	temp, ' '
	sbrc	flags, fCall
	ldi	temp, 'C'
	rcall	lcd_chr
	ldi	temp, ' '
	sbrc	flags, fCall
	ldi	temp, 'A'
	rcall	lcd_chr
	;current bit
	_lcd_goto 12, 0
	ldi	temp, '0'
	sbrc	flags, fDCFBit
	ldi	temp, '1'
	cpi	second, 0x59
	sbrc	flags, fLeapSecondNow
	cpi	second, 0x60
	brne	PC+2
	ldi	temp, '-'
	rcall	lcd_chr
	_lcd_chr ' '
	;second
	mov	temp, second
	rcall	lcd_bcd
	;(fall through to display_time)
;--------------------
display_time:
	;display date and time from Y (without modifying Y)
	_lcd_goto 0, 1
	;year
	_lcd_chr '2'
	_lcd_chr '0'
	ldd	temp, Y+tYear
	rcall	lcd_bcd
	_lcd_chr '-'
	;month
	ldd	temp, Y+tMonth
	rcall	lcd_bcd
	_lcd_chr '-'
	;day
	ldd	temp, Y+tDay
	rcall	lcd_bcd
	_lcd_chr ' '
	;hour
	ldd	temp, Y+tHour
	rcall	lcd_bcd
	_lcd_chr ':'
	;minute
	ldd	temp, Y+tMinute
	rjmp	lcd_bcd ;combined "rcall lcd_bcd" and "ret"
	
;--------------------

display_weekday:
	;display 2-digit weekday from time at Y on the LCD (without modifying Y)
	ldd	ZL, Y+tWeekday
	lsl	ZL
	subi	ZL, 2-weekdays
	rcall	ee_read
	rcall	lcd_chr
	rcall	ee_read
	rjmp	lcd_chr ;combined "rcall lcd_chr" and "ret"
	
;===============================================================================

dcf_output:
	;calculate current DCF77 output bit
	andi	flags, ~(1<<fDCFBit)
	mov	temp, second
	cpi	temp, 0x59
	sbrc	flags, fLeapSecondNow
	cpi	temp, 0x60
	brsh	dcf_output_end
	cpi	temp, 0x15
	brlo	dcf_output_found
	rcall	bcd2bin
	subi	temp, 15
	mov	YL, temp
	lsr	YL
	lsr	YL
	lsr	YL
	ld	temp2, Y
	andi	temp, 0x07
	breq	dcf_output_shift_end
dcf_output_shift_loop:
	lsr	temp2
	dec	temp
	brne	dcf_output_shift_loop
dcf_output_shift_end:
	sbrc	temp2, 0
	ori	flags, 1<<fDCFBit
dcf_output_found:
	ldi	temp, LOW(-410) ;4.194304 MHz / 1024 * 100 ms
	ldi	temp2, HIGH(-410)
	sbrs	flags, fDCFBit
	rjmp	dcf_output_set
	lsl	temp ;multiply time by 2 => 200 ms
	rol	temp2
dcf_output_set:
	out	TCNT1H, temp2
	out	TCNT1L, temp
	ldi	temp, 0x05 ;Clk/1024
	out	TCCR1B, temp
	sbi	PORTD, 5
	cbi	PORTD, 6
dcf_output_end:
	ret
	
;--------------------

update_dcftime:
	
	;copy current time/date to DCF77 Time
	ldi	YL, RAM_DCFTime
update_dcftime_copy:
	ldd	temp, Y+(RAM_Time-RAM_DCFTime)
	st	Y+, temp
	cpi	YL, RAM_DCFTime+7
	brne	update_dcftime_copy
	
	;increment DCF77 time by 1 minute
	ldi	YL, RAM_DCFTime
	rcall	inctime
	
	;generate new DCF77 frame
	
	;flags
	ldi	temp, 0x20<<2 ;start bit
	sbrc	flags, fCall
	ori	temp, 0x01<<2 ;Call bit
	sbrc	flags, fLeapSecond
	ori	temp, 0x10<<2 ;Leap Second bit
	lds	temp2, RAM_Time+tFlags
	sbrc	temp2, timeAnnounceDST ;use AnnounceDST flag from RAM_Time!
	ori	temp, 0x02<<2 ;Announce DST Switching bit
	ldi	YL, RAM_DCFTime ;TODO necessary? (still set from above?)
	ldd	temp2, Y+tFlags
	ori	temp, 0x08<<2 ;Winter Time bit
	sbrc	temp2, timeDST
	subi	temp, 0x04<<2 ;clear Winter Time bit, set Summer Time bit
	mov	dcf1, temp
	;minute
	ldd	temp, Y+tMinute
	rcall	dcf_parity_new
	sbrc	temp2, 0
	ori	temp, 0x20<<2 ;minute parity
	mov	dcf2, temp
	;hour
	ldd	temp, Y+tHour
	rcall	dcf_parity_new
	sbrc	temp2, 0
	ori	temp, 0x10<<2 ;hour parity
	mov	dcf3, temp
	;shift flags/minute/hour to correct position
	lsr	dcf3
	ror	dcf2
	ror	dcf1
	lsr	dcf3
	ror	dcf2
	ror	dcf1
	;day of month (bits 2~0 for DCF3)
	ldd	temp, Y+tDay
	rcall	dcf_parity_new
	andi	temp, 0x07
	lsl	temp
	swap	temp
	or	dcf3, temp ;DCF3: bits 7/6/5 = day bits 2/1/0
	;day of month (bits 5~3 for DCF4)
	ldd	temp, Y+tDay
	andi	temp, 0x07<<3 ;day bits 5/4/3 remain
	lsr	temp
	mov	dcf4, temp
	;day of week
	ldd	temp, Y+tWeekday
	rcall	dcf_parity
	swap	temp
	lsl	temp
	or	dcf4, temp ;weekday goes into DCF4 bits 7/6/5
	;month
	ldd	temp, Y+tMonth
	rcall	dcf_parity
	mov	dcf5, temp
	;shift day/weekday/month to correct position
	lsr	dcf5
	ror	dcf4
	lsr	dcf5
	ror	dcf4
	;year
	ldd	temp, Y+tYear
	rcall	dcf_parity
	andi	temp2, 0x01 ;date parity bit remains
	lsl	temp
	rol	temp2
	lsl	temp
	rol	temp2
	lsl	temp
	rol	temp2
	or	dcf5, temp
	mov	dcf6, temp2
	
	;leap second
	andi	flags, ~(1<<fLeapSecondNow)
	sbrs	flags, fLeapSecond
	rjmp	update_dcftime_leapsec_end
	ldd	temp, Y+tMinute
	cpi	temp, 0x00
	brne	update_dcftime_leapsec_end
	ori	flags, 1<<fLeapSecondNow
update_dcftime_leapsec_end:
	
	ret

;--------------------

dcf_parity_new:
	;clear temp2 before calculating parity
	clr	temp2
dcf_parity:
	;calculate parity bit (using temp2)
	push	temp
dcf_parity_loop:
	sbrc	temp, 0
	inc	temp2
	lsr	temp
	brne	dcf_parity_loop
	pop	temp
	ret
	